/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/btree.c $
 *
 * $Id: btree.c 105 2005-02-24 04:47:18Z kgreen $
 *
 ******************************************************************************
 */

#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif

static char *SVNID __attribute__((unused)) = "$Id: btree.c 105 2005-02-24 04:47:18Z kgreen $";

/*
 * Balanced binary tree utility routines.
 * 
 * The core algorithm is straight out of Donald Knuth's 
 * 'art of computer programming'.
 */

#include "btree.h"

/*
 * allocate and initialize a binary tree of void *objects which are
 * compared, created and deleted using the given methods.
 */


root_t btree_alloc (int  (*compare)(void *item1, void *item2),
                    void *(*create)(void *item), 
		    void  (*delete)(void *item))
{
  root_t root = (root_t) malloc(sizeof(struct root_s));

  if (NULL == root) return ((root_t) NULL);

  root->balanced = 1;
  root->depth = 0;
  root->compare = compare;
  root->create = create;
  root->delete = delete;
  root->node = NULL;

  return (root);
}


/*
 * An internal node deletion call back, used for by btree_free().
 */

static void _btree_free (node_t *node, void (*delete)(void *item))
{
  if (NULL != node->left) _btree_free(node->left, delete);
  if (NULL != node->right) _btree_free(node->right, delete);
  delete((void *) node->item);
  free(node);
  return;
}


/*
 * Walk a binary tree deleting all nodes.  
 * Then delete the tree root as well.
 */


int btree_free (root_t *root)
{
  if (NULL == *root) return (FALSE);
  if (NULL != (*root)->node) _btree_free((*root)->node, (*root)->delete);
  free (*root);
  *root = NULL;

  return (TRUE);
}


/*
 * find an item in a binary tree.
 */

void *btree_find (root_t root, void *item)
{
  node_t *node;
  node_t **child;
  int result;

  if (NULL == root) return((void *) NULL);

  child = &(root->node);

  while (NULL != (node = *child)) {
    result = root->compare(node->item, item);
    if (0 == result) return((void *) node->item);
    child = (result < 0) ? &node->left : &node->right;
  }
  return((void *) NULL);
}


/*
 * Internal routine to search for an item in a binary tree, 
 * optionally creating it if it is not found.  The tree is
 * rebalanced if necessary.
 *
 * This routine won't correctly rebalance the tree after any
 * deletion occurs, so a delete flag is maintained in the
 * tree root to determine if it's use is appropriate or not.
 */

static void *_btree_balanced_search (root_t root, void *item)
{
  node_t *r, *s, *t, *p, *q;
  node_t **pq;
  int a;
  int result;

  /*
   * no root - immediate error
   */

  if (NULL == root) return((void *) NULL);

  /*
   * special case - the tree is empty.  Always add the first node
   */

  if (NULL == root->node) {
    q = (node_t *) malloc(sizeof(struct node_s));
    if (NULL == q) return((void *) NULL);
    q->left = NULL;
    q->right = NULL;
    q->balance = 0;
    q->item = root->create(item);
    if (NULL == q->item) {
      free(q);
      return ((void *) NULL);
    }
    root->node = q;
    root->depth = 1;
    return ((void *) q->item);
  }

  /*
   * look for the item, keeping track of parent item 'p',
   * rebalancing point 's', address of link pointing to s, 'tp',
   * and address of the items' parent link 'qp', which 
   * points to 'q', the sought-after item.
   */

  t = NULL;
  s = p = root->node;

  while (1) {
    result = root->compare(p->item, item);
    if (0 == result) return((void *) p->item);
    pq = (result < 0) ? &(p->left) : &(p->right);
    if (NULL == (q = *pq)) break;
    if (0 != q->balance) {
      t = p;
      s = q;
    }      
    p = q;
  } 

  /*
   * the item was not found.  Allocate, initialize and link it in.
   */

  q = (node_t *) malloc(sizeof(struct node_s));
  if (NULL == q) return((void *) NULL);
  q->left = NULL;
  q->right = NULL;
  q->balance = 0;
  q->item = root->create(item);
  if (NULL == q->item) {
    free(q);
    return ((void *) NULL);
  }

  *pq = q;

  /*
   * adjust node balance factors between s and q to reflect the addition
   * of the new node.
   */

  result = root->compare(s->item, item);

  r = p = (result < 0) ? s->left : s->right;
  a = (result < 0) ? -1 : 1;

  while (p != q) {
    if (root->compare(p->item, item) < 0) {
      p->balance = -1;
      p = p->left;
    } else {
      p->balance = 1;
      p = p->right;
    }
  }

  /*
   * rebalance the tree.  There are three cases to consider
   */

  /*
   * the tree depth grew, but the tree remained in balance
   */

  if (0 == s->balance) {
    s->balance = a;
    root->depth += 1;
    return((void *) q->item);
  }

  /*
   * the tree became more balanced
   */

  if (-a == s->balance) {
    s->balance = 0;
    return((void *) q->item);
  }

  /*
   * the tree became less balanced
   */

#define LINK(a,p) *(((a) < 0) ? &((p)->left) : &((p)->right))

  result = r->balance;
  s->balance = r->balance = 0;
  p = r;

  if (-a == result) {

    /*
     * a double rotation must be performed
     */

    p = LINK(-a,r);
    LINK(-a,r) = LINK(a,p);
    LINK(a,p) = r;

    if (0 != p->balance) {
      if (a == p->balance) s->balance = -a;
      if (-a == p->balance) r->balance = a;
    }
  }

  /*
   * a single rotation must be performed
   */

  LINK(a,s) = LINK(-a,p);
  LINK(-a,p) = s;

#undef LINK

  /*
   * adjust link to new root node
   */

  if (NULL == t) {
    if (s == root->node)  root->node = p;
  } else {
    if (s == t->right) {
      t->right = p;
    } else {
      t->left = p;
    }
  }
  return((void *) q->item);
}


/*
 * Search for an item in a binary tree, creating it if it is not found.
 * 
 * Tree balance will be maintaind only as long as no deletions are made.
 */

void *btree_search (root_t root, void *item)
{
  node_t *node;
  node_t **child;
  int result;

  if (NULL == root) return((void *) NULL);
  child = &(root->node);

  /*
   * If the tree is still balanced, we can use the 
   * balanced version of this routine.  Otherwise
   * we just slip new objects in wherever they should go.
   */

  if (root->balanced) return(_btree_balanced_search(root, item));

  while (NULL != (node = *child)) {
    result = root->compare(node->item, item);
    if (0 == result) return((void *) node->item);
    child = (result < 0) ? &node->left : &node->right;
  }
  node = (node_t *) malloc(sizeof(struct node_s));
  if (NULL != node) {
    *child = node;
    node->left = NULL;
    node->right = NULL;
    node->item = root->create(item);
    if (NULL == node->item) {
      free(node);
      node = NULL;
    }
  }
  return((void *) (NULL == node ? NULL : node->item));
}


/*
 * Delete a specified item from a binary tree
 * The tree will become unbalanced, and the internal
 * routine that maintains balance will be disabled.
 */

void *btree_delete (root_t root, void *item)
{
  node_t *node;
  node_t **child;
  node_t *parent;
  node_t *l_sib;
  node_t *r_sib;
  int result;

  if (NULL == root) return((void *) NULL);

  parent = NULL;
  child = &(root->node);

  /*
   * look for the item, keeping track of parent node and the
   * address of the child pointer we followed from the parent
   */
  
  while (NULL != (node = *child)) {
    result = root->compare(node->item, item);
    if (0 == result) break;
    parent = node;
    child = (result < 0) ? &node->left : &node->right;
  } 
  if (NULL == node) return((void *) NULL);   /* item not found */
   
  root->balanced = 0;                        /* item found, disable balancing */

  /*
   * follow the links from the found node to locate a place we can
   * pick up and relink nodes subordinate to the found node.
   *
   * there are three cases
   */

  r_sib = node->right;

  /*
   * the trivial case where the found node has no children...
   */

  if (NULL == (l_sib = node->left)) l_sib = r_sib;
  else if (NULL != r_sib) {

    /*
     * ...one child...
     */

    if (NULL == r_sib->left) {
      r_sib->left = l_sib;
      l_sib = r_sib;
    } else {

      /*
       * ...and two children.  In this situation, we must find a child that
       * has at most one child of it's own.  That child replaces the deleted 
       * node and gains a new child of its own.
       */
      
      for (l_sib = r_sib->left; NULL != l_sib->left; l_sib = r_sib->left) {
        r_sib = l_sib;
      }
      r_sib->left = l_sib->right;
      l_sib->left = node->left;
      l_sib->right = node->right;
    }
  }
  root->delete(node->item);
  free(node);
  *child = l_sib;
  return((void *) (NULL == parent ? NULL : &parent->item));
}


/*
 * internal callback to walk a binary tree
 */

static void _btree_walk (node_t *node, 
                         void (*action)(void *item, void *data),
                         void *data)
{
  if (NULL != node->left) _btree_walk(node->left, action, data);
  action(node->item, data);
  if (NULL != node->right) _btree_walk(node->right, action, data);
  return;
}


/*
 * In-order visitation of every node of a binary tree.
 * the action routine is called for every item found.
 */

int btree_walk (root_t root, 
	        void (*action)(void *item, void *data),
		void *data)
{
  if ((NULL != root) && (NULL != root->node)) {
    _btree_walk(root->node, action, data);
    return(TRUE);
  }
  return(FALSE);
}


/*
 * A simple test driver to ensure correct operation,
 * not normally compiled.  Complile with -DTEST to 
 * generate a runnable test program.
 */


#ifdef TEST

typedef struct test_s {
  unsigned int created :1;
  unsigned int found   :1;
  unsigned int deleted :1;
  unsigned int missing :1;
} test_t;

#define NRAND 1000000
test_t value[NRAND];
int failed = 0;

int counter = 0;

static int compare (void *vitem1, void *vitem2)
{
  int *item1 = (int *) vitem1;
  int *item2 = (int *) vitem2;

  if (*item1 == *item2) return (0);
  return ((*item1 > *item2) ? -1 : 1);
}

static void *create (void *vitem)
{
  int *item = (int *) vitem;

  int *p;
  p = (int *) malloc(sizeof(int));
  *p = *item;
  return ((void *)p);
}

static void delete (void *vitem)
{
  int *item = (void *)vitem;

  free(item);
}

static void action (void *vitem, void *data)
{
  int *item = (int *)vitem;

  int i;
  
  counter +=1;
  
  if (value[*item].created) {
    value[*item].found = TRUE;
  } else {
    failed = 1;
    printf("Found %d, but it was never created!\n", *item);
  }
}

main ()
{
  root_t test;
  int *item;
  int i, j;
  int unique = 0;
  int deletions = 0;
  
  printf("Creating tree...\n");
  test = btree_alloc(compare, create, delete);
  if (NULL == test) {
    printf("Truly massive screwup.  Can't allocate tree root.\n");
    exit(TRUE);
  }
  
  for (i = 0; i<NRAND; i++) {
    value[i].created = FALSE;
    value[i].found = FALSE;
    value[i].deleted = FALSE;
    value[i].missing = FALSE;
  }
  
  printf("Populating a randomly generated tree...\n");
  
  for (i = 0; i<NRAND; i++) {
    j = rand() % NRAND;
    if (!value[j].created)  {
      unique += 1;
      value[j].created = TRUE;
    }
    item = btree_search(test, &j);
  }    
  
  printf("Created %d unique values out of a total of %d attempts.\n", unique, NRAND);
  printf("Walking tree to find them all...\n");
  
  btree_walk(test, action, (void *)0);
  
  for (i = 0; i<NRAND; i++) {
    if (value[i].created && !value[i].found) {
      failed = 1;
      printf("Item %d was created but is now missing from the tree.\n",i);
      value[i].missing = TRUE;
    }
    value[i].found = 0;
  }
  
  printf("Deleting random entries...\n");
  
  for (i = 0; i<NRAND; i++) {
    j = rand() % NRAND;
    if (value[j].created && !value[j].missing) {
      btree_delete(test, &j);
      value[j].deleted = TRUE;
      value[j].created = FALSE;
      deletions += 1;
      unique -=1;
    }
  }    
  
  printf("Deleted %d entries from the tree.  %d unique values remain.\n", deletions, unique);
  printf("Rechecking the remaining entries...\n");
  
  btree_walk(test, action, (void *)0);
  
  for (i = 0; i<NRAND; i++) {
    if (value[i].created && !value[i].found) {
      failed = 1;
      printf("Item %d is missing from the tree.\n",i);
      value[i].found = FALSE;
    }
  }
  
  printf("Creating additional random entries...\n");
  for (i = 0; i<NRAND; i++) {
    j = rand() % NRAND;
    if (!value[j].created) {
      btree_search(test, &j);
      value[j].deleted = FALSE;
      value[j].created = TRUE;
      unique += 1;
    }
  }    
  
  printf("There are now %d unique values in the tree.\n", unique);
  printf("Walking tree to find them all...\n");
  
  btree_walk(test, action, (void *)0);
  
  for (i = 0; i<NRAND; i++) {
    if (value[i].created && !value[i].found) {
      failed = 1;
      printf("Item %d was created but is now missing from the tree.\n",i);
      value[i].missing = 1;
    }
    value[i].found = FALSE;
  }
  
  printf("Deleting tree...\n");
  btree_free(&test);
  if (!failed) {
    printf("Done.  All tests passed.\n");
  } else {
    printf("Done.  At least one test failed.\n");
  }
}
#endif
