/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm.c $
 *
 * $Id: icm.c 117 2005-03-10 04:16:46Z kgreen $
 *
 ******************************************************************************
 */
  
#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif

static char *SVNID __attribute__((unused)) = "$Id: icm.c 117 2005-03-10 04:16:46Z kgreen $";

#include "icm-impl.h"
#include "dynstr.h"
#include "btree.h"

/*
 * utility functions
 */


icm_section_t *icm_find_section (icm_t *icm, char *name)
{
  if (icm && name) {
    char *temp = _icm_string(name);            /* make sure we are dealing with a lex string */
    icm_section_t **section = icm->sections;
    int count = icm->section_count;
    int i;
    for (i = 0; i < count; i++) {
      if (section[i]->name == temp) {
	section[i]->referenced = 1;
	return(section[i]);
      }
    }
  } 
  return ((icm_section_t *)0);
}


icm_pin_map_t *icm_find_pin_map (icm_family_t *family, char *name)
{
  if (family && name) {
    char *temp = _icm_string(name);            /* make sure we are dealing with a lex string */
    icm_pin_map_t **pinmap = family->pin_maps;
    int count = family->pin_map_count;
    int i;
    for (i = 0; i < count; i++) {
      if (pinmap[i]->name == temp) {
	pinmap[i]->referenced = 1;
	return(pinmap[i]);
      }
    }
  } 
  return ((icm_pin_map_t *)0);
}

icm_node_map_t *icm_find_node_map (icm_family_t *family, char *name)
{
  if (family && name) {
    char *temp = _icm_string(name);            /* make sure we are dealing with a lex string */
    icm_node_map_t **nodemaps = family->node_maps;
    int count = family->node_map_count;
    int i;
    for (i = 0; i < count; i++) {
      if (nodemaps[i]->name == temp) {
	nodemaps[i]->referenced = 1;
	return(nodemaps[i]);
      }
    }
  } 
  return ((icm_node_map_t *)0);
}

/*
 * The lexical analyzer's symbol table.
 * All syntactic elements returned to the parser are stored here.
 *
 * Only a single unique copy of any particular string is ever stored.
 * It is never deleted, thus the pointer itself serves as a unique id.
 *
 * The parser may freely dereference, keep, compare 
 * or discard pointers without fear of compromising memory.
 * It MAY NOT free them however.
 */

static int string_compare(void *item1, void *item2)
{
  return(strcmp((char *)item1, (char *)item2));
}

static void *string_create(void *item)
{
  return((void *)dyn_strdup((char *)item));
}

static void string_delete(void *item)
{
  free(item);
}

/*
 * Manage a btree of static strings
 */

static root_t string_table = (root_t)0; /* binary tree of ALL strings returned to the parser */

static void _icm_init_strings(void)
{
  if (string_table) btree_free(&string_table);          /* free any existing binary tree, deleting all members */
  
  string_table = btree_alloc(string_compare, string_create, string_delete);  /* allocate new binary tree */
  if (!string_table) {
    perror("fatal system error");
    abort();
  }
}

/*
 * parser string create one / delete all string(s)
 */

char *_icm_string (char *string)
{
  if (!string_table) _icm_init_strings();
  return (char *) btree_search(string_table, string);
}

/*
 * If you think you need to call _icm_free_string, 
 * I urge you to lie down and wait for the feeling to pass
 *
 * The icm_delete_icm() destructor will call it for you when the time is right
 */

static void _icm_free_strings(void)
{
  if (string_table) btree_free(&string_table);
  string_table = (root_t) 0;
}

/*
 * Parse an ICM file
 *
 * Arguments are passed to / recieved from yyparse via callbacks
 * There is literally no other way to do this without declaring yet
 * another batch of global values, which I greatly desire to not do.
 */

static void (*_icm_vmsg) PROTO((YYLTYPE *loc, level_t l, char *s, va_list ap)) = (void *) 0;
static char *ICM_version = (char *) 0;
static char *ICM_pathname = ""; 
static char *ICM_filename = (char *) 0;
static icm_t *ICM_result = (icm_t *) 0;
static int ICM_nerrors = 0;

int icm_parse (icm_t **icm, FILE *file, char *version, char *path, char *name, 
	       void (*msg_cback) PROTO((YYLTYPE *loc, level_t l, char *s, va_list ap)) )
{

  yyin = file;                         /* the passed FILE* must be OPEN and READY TO READ */

  ICM_version = version;               /* pass NULL to use current version, "" to honor [ICM Ver], string to force version */
  if (ICM_version) ICM_version = _icm_string(ICM_version);

  if (path) ICM_pathname = path;       /* pass NULL or "" to use the current working directory */
  if (ICM_pathname) ICM_pathname = _icm_string(ICM_pathname);

  ICM_filename = name;                 /* pass NULL when reading from a pipeline */
  if (ICM_filename) ICM_filename = _icm_string(ICM_filename);

  _icm_vmsg = msg_cback;               /* pass NULL to suppress the message function entirely */
  ICM_result = (icm_t *) 0;            /* parse result is stored here */
  ICM_nerrors = 0;                     /* zero error counter */

  yylex_init();                        /* setup the lexical analyzer */

  yyparse();                           /* Parse.  Callbacks are used to access/set the argument list */

  yylex_finish();                      /* Finish up.  Move along, not much to see here. */

  if (ICM_nerrors) {                          /* if there are parse errors */
    if (ICM_result) icm_free_icm(ICM_result); /* delete any partial icm struct the parser may have put together */
    ICM_result = (icm_t *) 0;                 /* and hand back NULL */
  }

  *icm = ICM_result;                   /* hand back the result, may be NULL */

  return(ICM_nerrors);                 /* return number of errors (0 == success) counted by _icm_msg() */

}

/*
 * Parser callbacks that utilize captured call arguments
 */

char *_icm_version (void)
{
  return(ICM_version);
}

void _icm_emit_result(icm_t *icm)
{
  ICM_result = icm;
}

void _icm_msg (level_t level, YYLTYPE *loc, char *s, ...)
{
  if (ERROR == level) ICM_nerrors += 1;  /* count errors for return status */
  if (_icm_vmsg) {                       /* no message hander?  We're done then. */
    va_list ap;
    va_start(ap, s);
    (*_icm_vmsg)(loc, level, s, ap);
    va_end(ap);
  }
}

/*
 * determine if an ICM specified  file exists.
 * automagically looks in the directory same directory as the .icm file
 *
 * returns  1 if the file does not exist in the same directory as the .icm file
 * returns  0 if the file does exist in the same directory as the .icm file
 * returns -1 if the file does not exist in the CWD, and we're reading from a pipeline.
 */

int _icm_file_exists (char *name)
{
  
  char *filename;
  size_t path_len;
  size_t name_len;
  char *pos;
  FILE *fp;
  
  if (!(name && strlen(name))) return 0;  /* no file name?  no problem! */
  
  path_len = strlen(ICM_pathname);
  name_len = strlen(name);
  
  filename = calloc(1, path_len + 1 + name_len + 1);
  pos = filename;
  
  if (path_len) {
    strncat(pos, ICM_pathname, path_len);
    pos += path_len;
    strncat(pos, SEPARATOR, 1);
    pos += 1;
  }
  strncat(pos, name, name_len); 
  
  fp = fopen(filename, "r");
  free((void *) filename);
  if (fp) {
    fclose(fp);
    return 0;
  }
  
  if (ICM_filename) {
    return 1;
  } else {
    return -1;
  }
}

char *_icm_file_name (void)
{
  return ICM_filename;
}

char *_icm_dir_name (void)
{
  return ICM_pathname;
}

/*
 * structure constructors/destructors
 *
 * Most of these are intended for internal use by the parser.
 *
 * Only the icm_free_icm() destructor is 'public' to provide
 * a clean way to deallocate icm_t structs returned to the
 * user by the parser.
 *
 * NOTE: Destructors DO NOT DELETE ANY ICM STRINGS
 * ICM STRINGS are 'owned' by the b-tree that holds them.
 *
 */

static int icm_ref_count = 0;  /* track number of icm_t structs handed out */

icm_t *_icm_new_icm (void)
{
  icm_ref_count += 1;
  return (icm_t *) calloc(1, sizeof(struct icm_s));
};

icm_header_t *_icm_new_icm_header (void)
{
  return (icm_header_t *) calloc(1, sizeof(struct icm_header_s));
};

icm_family_t *_icm_new_icm_family (void)
{
  return (icm_family_t *) calloc(1, sizeof(struct icm_family_s));
};

icm_model_t *_icm_new_icm_model (void)
{
  return (icm_model_t *) calloc(1, sizeof(struct icm_model_s));
};

icm_tree_path_t *_icm_new_icm_tree_path (void)
{
 return (icm_tree_path_t *) calloc(1, sizeof(struct icm_tree_path_s));
};

icm_node_path_t *_icm_new_icm_node_path (void)
{
  return (icm_node_path_t *) calloc(1, sizeof(struct icm_node_path_s));
};

icm_swath_entry_t *_icm_new_icm_swath_entry (void)
{
  return (icm_swath_entry_t *) calloc(1, sizeof(struct icm_swath_entry_s));
};

icm_pin_map_t *_icm_new_icm_pin_map (void)
{
  return (icm_pin_map_t *) calloc(1, sizeof(struct icm_pin_map_s));
};

icm_pin_list_entry_t *_icm_new_icm_pin_list_entry (void)
{
  return (icm_pin_list_entry_t *) calloc(1, sizeof(struct icm_pin_list_entry_s));
};

icm_node_map_t *_icm_new_icm_node_map (void)
{
  return (icm_node_map_t *) calloc(1, sizeof(struct icm_node_map_s));
};

icm_node_list_entry_t *_icm_new_icm_node_list_entry (void)
{
  return (icm_node_list_entry_t *) calloc(1, sizeof(struct icm_node_list_entry_s));
};

icm_section_t *_icm_new_icm_section (void)
{
  return (icm_section_t *) calloc(1, sizeof(struct icm_section_s));
};

icm_matrix_t *_icm_new_icm_matrix (void)
{
  return (icm_matrix_t *) calloc(1, sizeof(struct icm_matrix_s));
};

icm_matrix_value_t *_icm_new_icm_matrix_value (void)
{
  return (icm_matrix_value_t *) calloc(1, sizeof(struct icm_matrix_value_s));
};

icm_port_assignment_entry_t *_icm_new_icm_port_assignment_entry (void)
{
  return (icm_port_assignment_entry_t *) calloc(1, sizeof(struct icm_port_assignment_entry_s));
};

void icm_free_icm (icm_t *icm)
{
  int i;
  if (icm) {

    if (icm->header) _icm_free_icm_header(icm->header);

    if (icm->family) _icm_free_icm_family(icm->family);

    if (icm->sections) {
      for (i = 0; i < icm->section_count; i++) {
	_icm_free_icm_section(icm->sections[i]);
      }
      free((void *) icm->sections);  /* free the table too */
    }

    free((void *) icm);
    if (!(--icm_ref_count)) _icm_free_strings();  /* free strings if no remaining icm_t structs */
  }
};

void _icm_free_icm_header (icm_header_t *header)
{
  if (header) free((void *) header);
};

void _icm_free_icm_family (icm_family_t *family)
{
  int i;
  if (family) {

    if (family->models) {
      for (i = 0; i < family->model_count; i++) {
	_icm_free_icm_model(family->models[i]);
      }
      free((void *) family->models);  /* free the table too */
    }
 
    if (family->pin_maps) {
      for (i = 0; i < family->pin_map_count; i++) {
	_icm_free_icm_pin_map(family->pin_maps[i]);
      }
      free((void *) family->pin_maps);  /* free the table too */
    }

    if (family->node_maps) {
      for (i = 0; i < family->node_map_count; i++) {
	_icm_free_icm_node_map(family->node_maps[i]);
      }
      free((void *) family->node_maps);  /* free the table too */
    }

    free((void *) family);
  }
};

void _icm_free_icm_model (icm_model_t *model)
{
  int i;
  if (model) {

    if (model->tree_path) _icm_free_icm_tree_path(model->tree_path);

    if (model->node_path) {
      for (i = 0; i < model->node_path_count; i++) {
	_icm_free_icm_node_path(model->node_path[i]);
      }
      free((void *) model->node_path);  /* free the table too */
    }

    if (model->swath) {
      for (i = 0; i < model->swath_count; i++) {
	_icm_free_icm_swath_entry(model->swath[i]);
      }
      free((void *) model->swath);  /* free the table too */
    }

    free ((void *) model);
  }
};

void _icm_free_icm_tree_path (icm_tree_path_t *tree_path)
{
  if (tree_path) {
    if (tree_path->fork) _icm_free_icm_tree_path(tree_path->fork);
    if (tree_path->next) _icm_free_icm_tree_path(tree_path->next);
    free((void *) tree_path);
  }
};

void _icm_free_icm_node_path (icm_node_path_t *node_path)
{
  if (node_path) {
    if (node_path->nodes) free((void *) node_path->nodes);  /* free the table too */
    free((void *) node_path);
  }
};

void _icm_free_icm_swath_entry (icm_swath_entry_t *swath_entry)
{
  if (swath_entry) free((void *) swath_entry);
};

void _icm_free_icm_pin_map (icm_pin_map_t *pin_map)
{
  int i;
  if (pin_map) {
    if (pin_map->pin_list) {
      for (i = 0; i < pin_map->pin_list_count; i++) {
	_icm_free_icm_pin_list_entry(pin_map->pin_list[i]);
      }
      free((void *) pin_map->pin_list);  /* free the table too */
    }
  }
  free((void *) pin_map);
};

void _icm_free_icm_pin_list_entry (icm_pin_list_entry_t *pin_list_entry)
{
  if (pin_list_entry) free((void *) pin_list_entry);
};

void _icm_free_icm_node_map (icm_node_map_t *node_map)
{
  int i;
  if (node_map) {
    if (node_map->node_list) {
      for (i = 0; i < node_map->node_list_count; i++) {
	_icm_free_icm_node_list_entry(node_map->node_list[i]);
      }
      free((void *) node_map->node_list);  /* free the table too */
    }
  }
  free((void *) node_map);
};

void _icm_free_icm_node_list_entry (icm_node_list_entry_t *node_list_entry)
{
  if (node_list_entry) free((void *) node_list_entry);
};

void _icm_free_icm_section (icm_section_t *section)
{
  int i;
  if (section) {
    if (section->r) _icm_free_icm_matrix(section->r);
    if (section->l) _icm_free_icm_matrix(section->l);
    if (section->g) _icm_free_icm_matrix(section->g);
    if (section->c) _icm_free_icm_matrix(section->c);

    if (section->port_assignment) {
      for (i = 0; i < section->port_assignment_count; i++) {
	_icm_free_icm_port_assignment_entry(section->port_assignment[i]);
      }
      free((void *) section->port_assignment);  /* free the table too */
    }
    free((void *) section);
  }
};

void _icm_free_icm_matrix (icm_matrix_t *matrix)
{
  int i;
  if (matrix) {
    if (matrix->next) _icm_free_icm_matrix(matrix->next);
    if (matrix->values) {
      for (i = 0; i < matrix->value_count; i++) {
	_icm_free_icm_matrix_value(matrix->values[i]);
      }
      free((void *) matrix->values);  /* free the table too */
    }
    free((void *) matrix);
  }
};

void _icm_free_icm_matrix_value (icm_matrix_value_t *matrix_value)
{
  if (matrix_value) free((void **) matrix_value);
};

void _icm_free_icm_port_assignment_entry (icm_port_assignment_entry_t *port_assignment_entry)
{
  if (port_assignment_entry) free((void *) port_assignment_entry);
};

char *icm_vsprintf(char *f, va_list ap)
{

  /*
   * SAFELY use vsprintf to return a formatted string when
   * we don't know before hand how large the output string will turn
   * out to be.
   *
   * This is an ugly hack, but I know of no other way to do this
   * without risking buffer overflow.  The counted vsnsprintf is
   * not available under the C89 standard.  Even using vsnsprintf, 
   * two attempts are necessary to determine the actual buffer size 
   * and then format the string.
   *
   * This at least will always work, regardless of whether or not
   * vsnsprintf is available on any particular platform.
   *
   * This handles all format strings presently used by the parser, but
   * it's not a complete format scanner by any means.  If you add
   * a new parser message, you should make sure this routine
   * will actually handle it.
   */

  char *first, *last, *format=(char *)0, *string;
  char *text = (char *) calloc(1, sizeof(char));  /* start with a '\0' */

  int fudge=128; /* big enough to handle any numeric conversion */

  last = f;
  
  while (*last) {
    first = last;
    last = first + strcspn(first, "%");
    
    if (last - first) {

      /*
       * span simple character sequences, no conversion at all
       */

      format = (char *) calloc((last - first) + 1, sizeof(char));
      (void) strncpy(format, first, last - first);
      text = (char *) realloc((void *)text, strlen(text) + (last - first) + 1);
      sprintf(&text[strlen(text)], format);
      
    } else {

      /*
       * span a conversion specification
       * the conversion character is in last[-1]
       * the optional data modifier is in last [-2]
       * if absent last[-2] holds the '%' lead-in
       */

      if (first) {
	last = first + strspn(first, "%-+# 0123456789.hlL") + 1;
	format = (char *) calloc((last - first) + 1, sizeof(char));
	(void) strncpy(format, first, last - first);

	text = (char *) realloc((void *)text, strlen(text) + fudge + 1);  /* assume a numeric conversion */
	
	switch (last[-1]) {  /* what type of conversion is this? */
	  
	case 's' : /* a string - size it and allocate sufficient room */
	  string = va_arg(ap, char *);
	  text = (char *) realloc((void *)text, strlen(text) + strlen(string) + fudge + 1);
	  sprintf(&text[strlen(text)], format, string);
	  break;
	  
	case 'd' : /* various integer types */
	case 'i' :
	case 'o' :
	case 'x' :
	case 'X' :
	case 'u' :
	case 'c' :
	  switch (last[-2]) {
	  case 'l':
	    sprintf(&text[strlen(text)], format, va_arg(ap, long));
	    break;
	  case 'h' : /* fall through, shorts are promoted to int anyway */
	  default:
	    sprintf(&text[strlen(text)], format, va_arg(ap, int));
	    break;
	  }
	  break;
	  
	case 'f' :  /* various floating types */
	case 'e' :
	case 'E' :
	case 'g' :
	case 'G' :
	  switch (last[-2]) {
	  case 'L':
	    sprintf(&text[strlen(text)], format, va_arg(ap, long double));
	    break;
	  default:
	    sprintf(&text[strlen(text)], format, va_arg(ap, double));
	    break;
	  }
	  break;
	  
	  /*
	   * various pointer types
	   */
	  
	case 'p' :
	  sprintf(&text[strlen(text)], format, va_arg(ap, void *));
	  break;
	  
	case 'n' :
	  sprintf(&text[strlen(text)], format, va_arg(ap, int *));
	  break;
	  
	  /*
	   * print a bare %, no argument consumed at all
	   */
	  
	case '%' :
	  sprintf(&text[strlen(text)], format);
	  break;
	}
      }
    }
    if (format) {
      free((void *) format);
      format = (char *) 0;
    }
  }
  return (text);
}
