#ifndef ICM_H
#define ICM_H

/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icm.h $
 *
 * $Id: icm.h 105 2005-02-24 04:47:18Z kgreen $
 *
 ******************************************************************************
 */

/*
 * This is the public include file for the icm parser.
 * It is sufficient for accessing the parser public functions.
 */

#include "includes.h"

/*
 * Set package strings
 */

#define IBIS_ICM_VERSION PACKAGE_VERSION
#define IBIS_ICM_NAME    PACKAGE_NAME

/*
 * Some platform dependent stuff
 * 
 * File pathname separator character
 *
 * "%g" format specifier
 * microsoft-C interprets the field width of %g differently from most everyone else
 * and this messes up some regression tests.
 */

#ifdef _WIN32
#define SEPARATOR "\\"
#define G_FMT "%#.5g"
#else
#define SEPARATOR "/"
#define G_FMT "%#g"
#endif

/*
 * Parse error severity codes
 */

enum level_e {
  UNUSED = 0, INFO = 1, WARNING = 2, ERROR = 3, DEBUG = 4
};

typedef enum level_e level_t;

/*
 * Structure Definitions for the ICM parser
 */

typedef struct yyltype
{
  size_t first_line;
  size_t first_column;

  size_t last_line;
  size_t last_column;
} yyltype;

#define YYLTYPE yyltype  /* yylex requires a define, not a typedef */

/*
 * Define template structs so we don't have to order the actual definitions
 * from the bottom up. That makes them very hard to read
 */

typedef struct icm_s                       icm_t;
typedef struct icm_header_s                icm_header_t;
typedef struct icm_family_s                icm_family_t;
typedef struct icm_model_s                 icm_model_t;
typedef struct icm_tree_path_s             icm_tree_path_t;
typedef struct icm_node_path_s             icm_node_path_t;
typedef struct icm_swath_entry_s           icm_swath_entry_t;
typedef struct icm_pin_map_s               icm_pin_map_t;
typedef struct icm_pin_list_entry_s        icm_pin_list_entry_t;
typedef struct icm_node_map_s              icm_node_map_t;
typedef struct icm_node_list_entry_s       icm_node_list_entry_t;
typedef struct icm_section_s               icm_section_t;
typedef struct icm_matrix_s                icm_matrix_t;
typedef struct icm_matrix_value_s          icm_matrix_value_t;
typedef struct icm_port_assignment_entry_s icm_port_assignment_entry_t;

/*
 * A successful parse creates and populates an icm struct
 */

struct icm_s {
  YYLTYPE loc;
  int header_seen;
  icm_header_t *header;
  int family_seen;
  icm_family_t *family;
  int section_seen;
  int section_count;
  icm_section_t **sections;
};

struct icm_header_s {
  YYLTYPE loc;
  char *version;
  char *file_name;
  char *file_rev;
  char *date;
  char *source;
  char *notes;
  char *disclaimer;
  char *copyright;
  char *support;
  char *redist;
  char *redist_text;
};

struct icm_family_s {
  YYLTYPE loc;
  char *name;
  char *manufacturer;
  char *description;
  int model_seen;
  int model_list_seen;
  int model_count;
  icm_model_t **models;
  int pin_map_count;
  icm_pin_map_t **pin_maps;
  int node_map_count;
  icm_node_map_t **node_maps;
};

struct icm_model_s {
  YYLTYPE loc;
  char *name;
  char *mating;
  double slew;
  char *image;
  char *type;
  char *sgr;
  double ref_impedance;
  char *description;
  icm_tree_path_t *tree_path;
  int node_path_count;
  icm_node_path_t **node_path;
  int swath_present;
  int left_edge;
  int right_edge;
  int top_edge;
  int bottom_edge;
  int swath_count;
  int swath_rows;
  int swath_cols;
  icm_swath_entry_t **swath;
};

struct icm_tree_path_s {
  YYLTYPE loc;
  icm_tree_path_t *next;
  icm_tree_path_t *fork;
  char *type;
  char *name;
  char *side;
  char *scaletype;
  double scalefactor;
};

struct icm_node_path_s {
  YYLTYPE loc;
  char *type;
  char *name;
  char *side;
  char *scaletype;
  double scalefactor;
  int node_count;
  char **nodes;
};

struct icm_swath_entry_s {
  char *name;
  int row;
  int col;
};

struct icm_pin_map_s {
  YYLTYPE loc;
  icm_pin_map_t *next;
  char *name;
  int referenced;
  char *pin_order;
  int num_of_rows;
  int num_of_columns;
  int pin_list_count;
  icm_pin_list_entry_t **pin_list;
};

struct icm_pin_list_entry_s {
  char *pinname;
  char *bookname;
};

struct icm_node_map_s {
  YYLTYPE loc;
  char *name;
  int referenced;
  int node_list_count;
  icm_node_list_entry_t **node_list;
};

struct icm_node_list_entry_s {
  char *pinname;
  char *nodename;
  char *bookname;
};

struct icm_section_s {
  YYLTYPE loc;
  char *name;
  int referenced;
  char *derivation_method;
  int rows;
  int cols;
  icm_matrix_t *r;
  icm_matrix_t *l;
  icm_matrix_t *c;
  icm_matrix_t *g;
  int s_parameter_seen;
  char *s_file;
  int port_assignment_count;
  icm_port_assignment_entry_t **port_assignment;
};

struct icm_matrix_s {
  YYLTYPE loc;
  icm_matrix_t *next;
  char *type;
  char *representation;
  int bandwidth;
  int rows;
  int cols;
  double frequency;
  int value_count;
  icm_matrix_value_t **values;
};

struct icm_matrix_value_s {
  int row;
  int col;
  double value;
};

struct icm_port_assignment_entry_s {
  char *port;
  char *node;
};

/*
 * parser global utility functions
 */

int icm_parse PROTO((icm_t **icm, FILE *file, char *version, char *path, char *name, 
		     void (*msg_cback) PROTO((YYLTYPE *loc, level_t l, char *f, va_list ap)) ));
char *icm_vsprintf PROTO((char *f, va_list ap));

icm_section_t *icm_find_section PROTO((icm_t *icm, char *name));
icm_pin_map_t *icm_find_pin_map PROTO((icm_family_t *family, char *name));
icm_node_map_t *icm_find_node_map PROTO((icm_family_t *family, char *name));

void  icm_free_icm PROTO((icm_t *icm));

/*
 * macro definitions to aid in determining object types
 */

#define /* boolean */ IS_SLM(model)           (0 == strncmp((model)->type, "SLM_", 4))
#define /* boolean */ IS_MATED(model)         ((model)->mating == _icm_string("Mated")
#define /* boolean */ IS_SECTION(tree)        ((tree)->type == _icm_string("Section"))         
#define /* boolean */ IS_PINMAP(tree)         ((tree)->type == _icm_string("Model_pinmap"))
#define /* boolean */ IS_FORK(tree)           ((tree)->type == _icm_string("Fork"))
#define /* boolean */ IS_N_SECTION(node)      ((node)->type == _icm_string("N_section"))
#define /* boolean */ IS_MULT(tree_node)      ((tree_node)->scaletype == _icm_string("Mult"))
#define /* boolean */ IS_LEN(tree_node)       ((tree_node)->scaletype == _icm_string("Len"))
#define /* boolean */ IS_LUMPED(section)      ((section)->derivation_method == _icm_string("Lumped"))
#define /* boolean */ IS_DISTRIBUTED(section) ((section)->derivation_method == _icm_string("Distributed"))
#define /* boolean */ IS_NODEMAP(node)        ((node)->type == _icm_string("Model_nodemap"))
#define /* boolean */ IS_UNORDERED(pinmap)    ((pinmap)->pin_order == _icm_string("Unordered"))
#define /* boolean */ IS_ORDERED(pinmap)      ((pinmap)->pin_order != _icm_string("Unordered"))
#define /* boolean */ IS_S_PARAMETER(section) ((section)->s_parameter_seen || (0 != (section)->s_file))
#define /* boolean */ IS_RLGC(section)        (!(section)->s_parameter_seen && (0 == (section)->s_file))
#define /* boolean */ IS_DIAGONAL(matrix)     ((matrix)->representation == _icm_string("Diagonal_matrix"))

#define /* int     */ RLGC_TYPE(matrix)   ((_icm_string((matrix)->type) == _icm_string("[Resistance Matrix]")) ? RESISTANCE_MATRIX : ( \
                                          (_icm_string((matrix)->type) == _icm_string("[Inductance Matrix]")) ? INDUCTANCE_MATRIX : ( \
                                          (_icm_string((matrix)->type) == _icm_string("[Conductance Matrix]")) ? CONDUCTANCE_MATRIX : ( \
                                          (_icm_string((matrix)->type) == _icm_string("[Capacitance Matrix]")) ? CAPACITANCE_MATRIX : ( \
                                           0)))) \
					  )

#endif
