/*
 ******************************************************************************
 *
 * This file is a part of the IBIS ICM Parser.
 * Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 * 
 ******************************************************************************
 *
 * $URL: file:///home/kgreen/ibis-svn-root/icmchk/trunk/src/icmchk.c $
 *
 * $Id: icmchk.c 111 2005-02-25 06:07:45Z kgreen $
 *
 ******************************************************************************
 */

#if ! (defined(__GNUC__) || defined(__attribute__))
#define __attribute__(a) /* a */
#endif

static char *SVNID __attribute__((unused)) = "$Id: icmchk.c 111 2005-02-25 06:07:45Z kgreen $";

/*
 * Driver for the IBIS ICM parser
 */

char *copyright = \
"\
\n\
  This program is the IBIS ICM Parser version %s\n\
  Copyright (C) 2003-2005 Kelly C. Green <kgreen22@mindspring.com>\n\
\n\
  This program is free software; you can redistribute it and/or modify\n\
  it under the terms of the GNU General Public License as published by\n\
  the Free Software Foundation; either version 2 of the License, or\n\
  (at your option) any later version.\n\
\n\
  This program is distributed in the hope that it will be useful, but\n\
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY\n\
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License\n\
  for more details.\n\
\n\
  You should have received a copy of the GNU General Public License along\n\
  with this program; if not, write to the Free Software Foundation, Inc.,\n\
  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\
\n\
";

#include "icm.h"
#include "btree.h"
#include "dynstr.h"

/*
 * program name
 */

static char *ICM_progname = (char *)0;
static char ICM_default_version[4];

/*
 * error message statistics and verbosity control
 */

static char *level_names[5] = { "UNUSED", "INFO", "WARNING", "ERROR", "DEBUG" }; 
static int counters[5] = {0, 0, 0, 0, 0};

static int verbosity = 0;
static int verbosity_table[4][5] = { {0, 0, 0, 0, 0},
                                     {0, 0, 1, 1, 0},
                                     {0, 1, 1, 1, 0},
                                     {0, 1, 1, 1, 1}};

/*
 * Error messages are optionally stored in a binary tree, 
 * ordered by line number.  They are all printed out in order 
 * after the parse is complete.
 */

typedef struct message_s message_t;

struct message_s {
  YYLTYPE loc;
  int count;
  level_t level;
  char *text;
};

/*
 * The basic error report
 */

static void report (message_t *msg)
{
  fprintf(stderr, "%s (Line %i/Col %i) %s\n", level_names[msg->level], msg->loc.first_line, msg->loc.first_column, msg->text);
}

/*
 * If error messages are to be sorted, they are stored in a btree
 * The following establishes the btree and it's handlers.
 */

static root_t message_table = (root_t) 0;

static int message_compare (void *item1, void *item2)
{
  message_t *message_1 = (message_t *) item1;
  message_t *message_2 = (message_t *) item2;

  /*
   * sort by line number first
   */
  
  if (message_1->loc.first_line != message_2->loc.first_line) {
    return((int) (message_2->loc.first_line - message_1->loc.first_line));
  }

  /*
   * same line number - sort by column number
   */
  
  if (message_1->loc.first_column != message_2->loc.first_column) {
    return((int) (message_2->loc.first_column - message_1->loc.first_column));
  }
  
  /*
   * same line/col number, sort by message level
   */
  
  if (message_1->level != message_2->level) {
    return(message_2->level - message_1->level);
  }
  
  /*
   * same line/column number and severity - sort by message text
   */
  
  return (dyn_strcasecmp(message_2->text, message_1->text));
}

static void *message_create (void *item)
{
  message_t *message = (message_t *)item;
  message_t *new_message = (message_t *) calloc(1, sizeof(message_t));
  
  new_message->level = message->level;
  new_message->loc = message->loc;
  new_message->text = strdup(message->text);
  return (void *) new_message;
}

static void message_delete (void *item)
{
  message_t *message = (message_t *)item;
  free((void *) message->text);
  free((void *) message);
}

static void message_print (void *message, void *data)
{
  message_t *msg = (message_t *)message;
  int i;
  for (i = msg->count; i > 0; i--) report(msg);
}


/*
 * By design, the parser does not directly emit messages,
 * but depends on a callback to do whatever is desired.
 *
 * This is intended to facilitate integration with an eda
 * tool.
 *
 * The stand-alone icmchk driver provides the following
 * routine as an example of capturing and reporting errors to stderr.
 */

static void message_cback (YYLTYPE *loc, level_t l, char *f, va_list ap)
{
  message_t *msg;
  
  counters[l] += 1;
  if (verbosity_table[verbosity][l]) {
    
    msg = (message_t *) calloc(1, sizeof(message_t));
    
    if (loc) {
      msg->loc = *loc;
    }
    
    msg->level = l;
    msg->text = icm_vsprintf(f, ap); /* allocates return buffer, so it won't overflow, ever */

    /*
     * having formed a message block, we now either sort and store
     * it, or just print it immediately, depending on options.
     */
    
    if (message_table) {
      msg = btree_search(message_table, (void *) msg);
      msg->count += 1;  /* count references in case of duplicates */
    } else {
      report(msg);
      free((void *) msg->text);
      free((void *) msg);
    }
  }
}

/*
 * A friendly usage message
 */

static void usage (void)
{
  char *version;
  char *last;

  /*
   * drop off the last digit from the XX.YY.ZZ string
   */

  version = strdup(IBIS_ICM_VERSION);
  for (last = &version[strlen(version)]; version < last; last--) if ('.' == *last) break;
  if ('.' == *last) *last = '\0';

  fprintf(stdout,"\n\
IBIS ICM Version %s parser.\n\n\
USAGE: %s [-vschp] [-f[<ver>]] [<file name>]\n\
       -v --> (verbose, once)  Report additional informational messages.\n\
       -v --> (verbose, twice) Report debug messages (caution: voluminous).\n\
       -s --> (sort)           Sort diagnostic messages by line number. \n\
       -c --> (copyright)      Display program copyright and exit.\n\
       -h --> (help)           Display this message and exit.\n\
       -p --> (pipe)           Read stdin pipeline instead of file.\n\
       -f --> (force version)  If -f<ver>, parse as version '<ver>'.\n\
                               If -f, honor the file's [ICM Ver] keyword.\n\
                               If omitted, always parse as version '%s'. \n\
       <file name>             The input IBIS ICM Model file.\n\n",
  IBIS_ICM_VERSION, ICM_progname, version);

  free((void *) version);
}

/*
 * The stand-alone icmchk driver
 */

int main (int argc, char **argv)
{
  char *sep;
  size_t path_len;
  int pipe = 0;
  
  char *ICM_filename = (char *) NULL;
  char *ICM_pathname = "";

  char *ICM_version = (char *) NULL;

  icm_t *icm = (icm_t *) NULL;  /* The ICM parse resolves into this structure */
  
  verbosity = 1;                /* set minimum verbosity (warnings) */

  /*
   * Extract the program name for usage()
   */

  sep = strrchr(argv[0], SEPARATOR[0]);
  if (!sep) {
    ICM_progname = argv[0];
  } else {
    sep[0] = '\0';
    ICM_progname = &sep[1];
  }

  /*
   * Extract the program default version for usage() and -f<ver>
   */

  strncpy(ICM_default_version, IBIS_ICM_VERSION, 3);
  ICM_default_version[3] = '\0';

  /*
   * I'd love to use getopt(), but it's not available on windows
   * Thus, parse options the hard way
   */

  while ( --argc > 0 ) {
    size_t n = 1;                  /* process concatenated options */
    argv++;
    switch ((*argv)[0]) {
    case '-':  
      while ((*argv)[n]) {
	switch ((*argv)[n++]) {
	case 'v': 
	  verbosity += 1;
	  if (verbosity > 3) verbosity = 3;
	  break;
	case 'c': 
	  fprintf(stdout, copyright, IBIS_ICM_VERSION);
	  exit(0);
	case 's': 
	  message_table = btree_alloc(message_compare, message_create, message_delete);
	  break;
	case 'p':
	  pipe = 1;
	  break;
	case 'h':
	  usage();
	  exit(0);
	  break;
	case 'f':                                                  /* gather up an option argument */
	  if ((*argv)[n]) {                                        /* check for a concatenated argument */
	    ICM_version = &(*argv)[n];                             /* if so, consume what's left of the argument */
            n = strlen(*argv);
	  } else {                                                 /* no argument given */
	    ICM_version = ICM_default_version;                     /* use the parser's default */
	  }
	  break;
	default:    /* unrecognized option of some sort */
	  usage();
	  exit(1);
	}
      }
      break;
    default:        /* It's not an option, so it's got to be a filename */
      
      /*
       * We can't have both a pipe and a filename, or more than one filename
       */
      
      if (ICM_filename || pipe) {
	usage();
	exit(1);
      }

      if (!freopen( argv[0], "r", stdin)) {
	perror(argv[0]);
	exit(1);
      }

      /*
       * Strip the file name from the path name.
       *
       * We need the path to support a check for file existance 
       * in any file references within the given ICM file.
       *
       * The path separator is OS dependent.
       */
      
      sep = strrchr(argv[0], SEPARATOR[0]);
      if (!sep) {
	ICM_filename = argv[0];
	ICM_pathname = "";
      } else {
	sep[0] = '\0';
	path_len = strlen(argv[0]);
	ICM_filename = &sep[1];
	ICM_pathname = (char *) calloc(1, path_len + 1 + 1);
	strncat(ICM_pathname, argv[0], path_len);
	ICM_pathname[path_len] = SEPARATOR[0];
	ICM_pathname[path_len + 1] = '\0';
      }
    } 
  }

  /*
   * must have either a pipe or a filename
   */

  if (!(pipe || ICM_filename)) {
    usage();
    exit(1);
  }

  /*
   * Identify self, and invoke the parser.
   */

  fprintf(stdout, "IBIS ICM Parser, version %s\n", IBIS_ICM_VERSION);

  fflush(stdout);

  /*
   * The arguments to the parser are as follows:
   *
   * &icm        - The location to store the pointer to the resulting struct_s icm
   *
   * stdin       - The open FILE * to parse
   *
   * ICM_VERSION - NULL for a default v 1.1 parse
   *               an empty string "" to honor the file's [ICM Ver] keyword
   *               a specific version "1.0" or "1.1" to determine the specific version to use
   *
   * ICM_pathname - The directory path leading to the parsed .icm file
   *                Use an empty string "" if parsing a pipeline.
   *
   * ICM_filename - The name of the .icm file.  Use NULL if parsing a pipeline.
   *
   * message_cback - A function called by the parser to handle error messages
   *                 as they occur.  It is called as follows:  
   *
   *                 message_cback (YYLTYPE *err_location, level_t error_level, char *format, va_list ap)
   *
   *                 Passing NULL, will suppress all error callbacks, and thus
   *                 all error messages.
   *
   * return value:  True if parse errors were encountered, false otherwise.
   *                or, just look to see if you got an icm struct.
   *                &icm is set to NULL if there are any errors.
   *
   */

  (void) icm_parse(&icm, stdin, ICM_version, ICM_pathname, ICM_filename, message_cback);

  /*
   * If no parse errors were detected, then the icm struct handed back is
   * a fully resolved binary representation of the input file, omitting
   * only comments and ignorable text outside the boundaries of the
   * ICM begin/end keyword pair.
   *
   * NOTE: Unless you enjoy memory leaks and exciting application behavior,
   * treat the icm structure as 'read-only'.  The parser owns the
   * struct, and all pointers within it.  icm_parse is a constructor.  
   * Release unwanted icm structs by using the corresponding destructor 
   * (see example below) when you are done with it.
   *
   * You may keep the structure as long as you want.  If you parse
   * additional files in the meantime, you will wind up holding multiple 
   * icm structs.
   *
   * See icm.h for the content of the struct, and a small collection of
   * utility routines for poking about in it.  If more detail is needed
   * read the code in icm-yacc.y.  There are many validation routines
   * toward the end of the module that munge around with the struct
   * in various interesting ways.
   *
   * The same structure is returned for both ICM 1.0 and 1.1 files.
   *
   * NOTE: Don't assume that binary compatibility can be maintained 
   * for all future versions of the ICM specification.  We don't know what
   * future versions of the ICM specification will bring.
   */

  /*
   * If messages were sorted and stored for later printout, the
   * time is now.  Print them all at once and then invoke the destructor.
   */

  if (message_table) {
    btree_walk(message_table, message_print, (void *) NULL);
    btree_free(&message_table);
  }

  fflush(stderr);

  /*
   * Report overall parse status
   */

  fprintf(stdout, "Finished. ERRORS: %d", counters[ERROR]);
  if (verbosity >=1) fprintf(stdout, " WARNINGS: %d", counters[WARNING]);
  if (verbosity >=2) fprintf(stdout, " INFORMATIONAL: %d", counters[INFO]);
  if (verbosity >=3) fprintf(stdout, " DEBUG: %d", counters[DEBUG]);
  fprintf(stdout, "\n");

  /*
   * icm_parse & icm_free_icm are a constructor/destructor pair.
   *
   * The destructor recognizes when all outstanding icm structs
   * have been returned, and then performs additional internal
   * cleanups within the parser that can't be done as long as
   * any icm structs exist.
   */

  if (icm) icm_free_icm(icm);

  /*
   * Having freed the only outstanding icm struct, all parser memory
   * has now been deallocated.
   *
   * The parser should have NO memory leaks.
   * Yes, I did check this fairly carefully.
   */

  /*
   * Exit with error if ERROR count is nonzero
   */

  if (ICM_filename && fclose(stdin)) {
    perror(ICM_filename);
  }
  
  exit (0 != counters[ERROR]);
}

